#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <time.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>
#include <linux/dvb/audio.h>
#include <linux/dvb/version.h>
#ifdef B25
#include "arib_std_b25.h"
#include "b_cas_card.h"
#else
typedef void ARIB_STD_B25;
#endif

#define FLAG_VERBOSE		0x0001
#define TUNING_TIMEOUT		5
#define DVR_BUFFER_SIZE		(188 * 44 * 1024)	/* ~= 8MB */
#define DMX_BUFFER_SIZE		(2 * 1024 * 1024)	/* 2MB */
#define IO_BUFFER_SIZE		(188 * 256)		/* ~= 48K */
#define DMX_MAX_FILTERS		8
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
	int				adapter_nr;
	int				flags;
	int				fd_fe;
	int				fd_dmx[DMX_MAX_FILTERS];
	int				dvr_buffer_size;
	int				dmx_buffer_size;
	ARIB_STD_B25			*b25;
	struct dvb_frontend_info	fe_info;
	struct dvb_frontend_parameters	fe_param;
	struct dmx_pes_filter_params	dmx_filter_param[DMX_MAX_FILTERS];
	char				fe_dev[32];
	char				dvr_dev[32];
	char				dmx_dev[32];
} dvb_control_t;

/* signal */
static int signal_flag = 0;
void signal_handler(int arg)
{
	signal_flag = arg;
}

#ifdef B25
/* Initialize ARIB B25 library */
static ARIB_STD_B25 *b25_init(int round, int strip, int emm)
{
	int ret;
	ARIB_STD_B25 *b25;
	B_CAS_CARD *bcas;

	b25 = create_arib_std_b25();
	if ( b25 == NULL ) {
		fprintf(stderr, "failed to create arib_std_b25\n");
		return NULL;
	}
	ret = b25->set_multi2_round(b25, round);
	if ( ret < 0 ) {
		fprintf(stderr, "b25->set_multi2_round(%d) failed - %d\n", round, ret);
		goto err;
	}
	ret = b25->set_strip(b25, strip);
	if ( ret < 0 ) {
		fprintf(stderr, "b25->set_strip(%d) failed - %d\n", strip, ret);
		goto err;
	}
	ret = b25->set_emm_proc(b25, emm);
	if ( ret < 0 ) {
		fprintf(stderr, "b25->set_emm_proc(%d) failed - %d\n", emm, ret);
		goto err;
	}

	bcas = create_b_cas_card();
	if ( bcas == NULL ) {
		fprintf(stderr, "failed to create b_cas_card\n");
		goto err;
	}
	ret = bcas->init(bcas);
	if ( ret < 0 ) {
		fprintf(stderr, "bcas->init failed - %d\n", ret);
		bcas->release(bcas);
		goto err;
	}
	ret = b25->set_b_cas_card(b25, bcas);
	if ( ret < 0 ) {
		fprintf(stderr, "b25->set_b_cas_card failed - %d\n", ret);
		bcas->release(bcas);
		goto err;
	}

	return b25;

err:
	b25->release(b25);
	return NULL;
}
#endif

/* Tune in the specified frequency */
static int tune(dvb_control_t *dvb, int freq)
{
	fe_status_t status;
	time_t time_timeout, time_now;
	struct pollfd pfd[1];
	struct dvb_frontend_event event;

	if ( dvb->fd_fe <= 0 ) {
		/* Open frontend */
		dvb->fd_fe = open(dvb->fe_dev, O_RDWR);
		if ( dvb->fd_fe < 0 ) {
			perror(dvb->fe_dev);
			return EXIT_FAILURE;
		}

		/* Check if the adapter is suitable */
		if ( ioctl(dvb->fd_fe, FE_GET_INFO, &dvb->fe_info) < 0 ) {
			perror("ioctl FE_GET_INFO");
			goto err;
		}
		if ( dvb->fe_info.type != FE_OFDM ) {
			fprintf(stderr, "%s: unknown type of adapter\n", dvb->fe_info.name);
			goto err;
		}
		if ( dvb->flags & FLAG_VERBOSE ) {
			fprintf(stderr, "Open frontend: %s\n", dvb->fe_info.name);
		}

		/* Get initial FE parameters */
		if ( ioctl(dvb->fd_fe, FE_GET_FRONTEND, &dvb->fe_param) < 0 ) {
			perror("ioctl FE_GET_FRONTEND");
			goto err;
		}
	}

	/* Set frequency */
	dvb->fe_param.frequency = (uint32_t)freq * 1000000 + (1000000 / 7);
	if ( dvb->fe_param.frequency < dvb->fe_info.frequency_min
	  || dvb->fe_param.frequency > dvb->fe_info.frequency_max ) {
		fprintf(stderr, "The frequency (%dMHz) is out of range\n", freq);
		/* do not close */
		return EXIT_FAILURE;
	}
	if ( ioctl(dvb->fd_fe, FE_SET_FRONTEND, &dvb->fe_param) < 0 ) {
		perror("ioctl FE_SET_FRONTEND");
		goto err;
	}
	/* Wait for asynchronous event */
	pfd[0].fd = dvb->fd_fe;
	pfd[0].events = POLLIN;
	if ( poll(pfd, 1, 3000) ) {
		if ( pfd[0].revents & POLLIN ) {
			if ( ioctl(dvb->fd_fe, FE_GET_EVENT, &event) == -EOVERFLOW ) {
				perror("ioctl FE_GET_EVENT");
				goto err;
			}
			if ( event.parameters.frequency <= 0 ) {
				fprintf(stderr, "Failed to set frequency at %u Hz  (status %02x).\n",
					dvb->fe_param.frequency, event.status);
				goto err;
			}
		}
	}

	/* Wait for tuning */
	time_timeout = time(NULL);
	time_timeout += TUNING_TIMEOUT;
	status = 0;
	do {
		usleep(50 * 1000);
		if ( ioctl(dvb->fd_fe, FE_READ_STATUS, &status) < 0 ) {
			perror("ioctl FE_READ_STATUS");
			goto err;
		}
		if ( status & FE_HAS_LOCK ) {
			break;
		}

		/* Prevent a hang:
		 * FE_TIMEOUT does not happen in no signal condition.
		 */
		time_now = time(NULL);
		if ( time_now > time_timeout ) {
			break;
		}
	} while ( !(status & FE_TIMEDOUT) );

	if ( !(status & FE_HAS_LOCK) ) {
		fprintf(stderr, "Failed to tune at %u Hz (status %02x).\n",
			dvb->fe_param.frequency, status);
		/* do not close */
		return EXIT_FAILURE;
	}

	if ( dvb->flags & FLAG_VERBOSE ) {
		uint16_t snr;

		fprintf(stderr, "Tuned at %uHz", dvb->fe_param.frequency);
		if ( ioctl(dvb->fd_fe, FE_READ_SNR, &snr) < 0 ) {
			perror("ioctl FE_READ_STATUS");
			/* Ignore error */
		}
		else {
			fprintf(stderr, "  S/N: %d.%ddB",
				(snr >> 8), (((snr & 0xFF) * 10) >> 8));
		}
		fputc('\n', stderr);
	}

	return 0;

err:
	close(dvb->fd_fe);
	dvb->fd_fe = 0;
	return EXIT_FAILURE;
}

/* Route all the DVB TS data to dvr device */
static int dmx_route_dvr(dvb_control_t *dvb)
{
	struct dmx_pes_filter_params filter;

	if ( dvb->fd_dmx[0] <= 0 ) {
		/* Open Demux */
		dvb->fd_dmx[0] = open(dvb->dmx_dev, O_RDWR);
		if ( dvb->fd_dmx[0] < 0 ) {
			perror(dvb->dmx_dev);
			return EXIT_FAILURE;
		}
	}
	else {
		if ( ioctl(dvb->fd_dmx[0], DMX_STOP) < 0 ) {
			perror("ioctl DMX_STOP");
			return EXIT_FAILURE;
		}
	}

	if ( dvb->dmx_buffer_size > 0 ) {
		if ( ioctl(dvb->fd_dmx[0], DMX_SET_BUFFER_SIZE, dvb->dmx_buffer_size) < 0 ) {
			perror("ioctl DMX_SET_BUFFER_SIZE");
			/* Ignore error & continue */
		}
	}

	/* route all PIDs to vdr */
	filter.pid = 0x2000;	/* treat as all PIDs */
	filter.input = DMX_IN_FRONTEND;
	filter.output = DMX_OUT_TS_TAP;
	filter.pes_type = DMX_PES_OTHER;
	filter.flags = DMX_IMMEDIATE_START;

	if ( ioctl(dvb->fd_dmx[0], DMX_SET_PES_FILTER, &filter) < 0 ) {
		perror("ioctl DMX_SET_PES_FILTER");
		close(dvb->fd_dmx[0]);
		dvb->fd_dmx[0] = 0;
		return EXIT_FAILURE;
	}

	return 0;
}

/* Record TS data on the specified file */
static int record_dvr(dvb_control_t *dvb, char *file, int duration, size_t length)
{
	int ret, fd_dvr;
	FILE *fp_out;
	size_t xlen;
	uint8_t *buff, *data;
#ifdef B25
	ARIB_STD_B25 *b25 = dvb->b25;
	int b25_disable = FALSE;
	ARIB_STD_B25_BUFFER b25_in, b25_out;
#endif
	struct sigaction sa;
	struct pollfd pfd[1];

	buff = malloc(IO_BUFFER_SIZE);
	if ( buff == NULL ) {
		perror("IO_BUFFER");
		return EXIT_FAILURE;
	}

	/* Open DVR */
	fd_dvr = open(dvb->dvr_dev, O_RDONLY|O_NONBLOCK);
	if ( fd_dvr < 0 ) {
		perror(dvb->dvr_dev);
		return EXIT_FAILURE;
	}

	if ( strcmp(file, "-") == 0 ) {
		fp_out = stdout;
	}
	else {
		fp_out = fopen(file, "w");
		if ( fp_out == NULL ) {
			perror(file);
			close(fd_dvr);
			return EXIT_FAILURE;
		}
	}

	/* Set signal hander */
	memset(&sa, 0, sizeof(struct sigaction));
	sa.sa_handler = signal_handler;
	sa.sa_flags = SA_RESTART;
	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);
	sigaction(SIGALRM, &sa, NULL);

#ifdef B25
	b25_in.data = buff;
	b25_in.size = 0;
	b25_out.data = NULL;
	b25_out.size = 0;
#endif
	pfd[0].fd = fd_dvr;
	pfd[0].events = POLLIN;

	if ( dvb->flags & FLAG_VERBOSE ) {
		time_t t;
		t = time(NULL);
		fprintf(stderr, "Start recording at %s", ctime(&t));
	}
	if ( duration > 0 ) {
		alarm(duration);
	}

	/* recorder main */
	ret = 0;
	xlen = 0;
	do {
		if ( poll(pfd, 1, 1) ) {
			if ( pfd[0].revents & POLLIN ) {
				ssize_t n;
				n = read(fd_dvr, buff, IO_BUFFER_SIZE);
				if ( n < 0 ) {
					perror(dvb->dvr_dev);
					ret = EXIT_FAILURE;
					break;
				}
				data = buff;
#ifdef B25
				if ( n > 0 && !b25_disable && b25 != NULL ) {
					int code;

					b25_in.size = n;
					code = b25->put(b25, &b25_in);
					if ( code < 0 ) {
						fprintf(stderr, "b25->put() failed - %d\n", ret);
						/* continue without B25 */
						b25_disable = TRUE;
					}
					else {
						code = b25->get(b25, &b25_out);
						if ( code < 0 ) {
							fprintf(stderr, "b25->get() failed - %d\n", ret);
							/* continue without B25 */
							b25_disable = TRUE;
						}
						else {
							data = b25_out.data;
							n = b25_out.size;
						}
					}
				}
#endif	/* B25 */
				if ( n > 0 ) {
					size_t w;
					w = fwrite(data, 1, n, fp_out);
					xlen += w;
					if ( w != n ) {
						perror(file);
						ret = EXIT_FAILURE;
						break;
					}
				}
			}
		}
	} while ( !signal_flag && (length <= 0 || xlen < length) );

	close(fd_dvr);

	/* Reset signal hander */
	memset(&sa, 0, sizeof(struct sigaction));
	sa.sa_handler = SIG_DFL;
	sigaction(SIGALRM, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

#ifdef B25
	/* process remaining data in ARIB B25 buffer */
	if ( b25 != NULL && ret == 0 ) {
		int code;
		code = b25->flush(b25);
		if ( code < 0 ) {
			fprintf(stderr, "b25->flush() failed - %d\n", ret);
			b25_disable = TRUE;
		}
		if ( !b25_disable && signal_flag != SIGPIPE ) {
			code = b25->get(b25, &b25_out);
			if ( code < 0 ) {
				fprintf(stderr, "b25->get() failed - %d\n", ret);
			}
			else {
				ssize_t n;
				data = b25_out.data;
				n = b25_out.size;
				if ( n > 0 && xlen < length ) {
					size_t w;
					w = fwrite(data, 1, n, fp_out);
					xlen += w;
					if ( w != n ) {
						perror(file);
						ret = EXIT_FAILURE;
					}
				}
			}
		}
	}
#endif	/* B25 */

	if ( fclose(fp_out) == EOF ) {
		perror(file);
		ret = EXIT_FAILURE;
	}
	if ( dvb->flags & FLAG_VERBOSE ) {
		time_t t;
		t = time(NULL);
		if ( signal_flag && signal_flag != SIGALRM ) {
			psignal(signal_flag, "Terminated by signal");
		}
		fprintf(stderr, "End recording at %s", ctime(&t));
	}
#ifdef B25
	if ( b25 != NULL ) {
		if ( dvb->flags & FLAG_VERBOSE ) {
			int i, n, code;
			ARIB_STD_B25_PROGRAM_INFO info;

			n = b25->get_program_count(b25);
			if ( n > 0 ) {
				fprintf(stderr, "    Prog     total TS  undecrypt TS  unpurchased  error\n");
			}
			for ( i = 0; i < n; i++ ) {
				fprintf(stderr, "%2d ", i + 1);
				code = b25->get_program_info(b25, &info, i);
				if ( code < 0 ) {
					putc('\n', stderr);
					continue;
				}
				fprintf(stderr, "%5d %12"PRId64"  %12"PRId64"   %10d   %04x\n",
					info.program_number,
					info.total_packet_count,
					info.undecrypted_packet_count,
					info.ecm_unpurchased_count,
					info.last_ecm_error_code);
			}
		}
		b25->reset(b25);
	}
#endif	/* B25 */

	return ret;
}

static void dvb_shutdown(dvb_control_t *dvb)
{
	int i;

	for ( i = 0; i < DMX_MAX_FILTERS; i++ ) {
		if ( dvb->fd_dmx[i] > 0 ) {
			close(dvb->fd_dmx[i]);
		}
	}
	close(dvb->fd_fe);

#ifdef B25
	if ( dvb->b25 != NULL ) {
		dvb->b25->release(dvb->b25);
	}
#endif	/* B25 */
}

/* Convert channel number to frequency in MHz */
static int channel2freq(int channel, int is_catv)
{
	int freq = 0;

	if ( is_catv ) {
		if ( channel < 13 || channel > 63 ) {
			freq = 0;
		}
		else if ( channel < 22 ) {
			freq = 111 + (channel - 13) * 6;
		}
		else if ( channel == 22 ) {
			freq = 167;
		}
		else if ( channel == 23 ) {
			freq = 225;
		}
		else if ( channel < 28 ) {
			freq = 233 + (channel - 24) * 6;
		}
		else {
			freq = 255 + (channel - 28) * 6;
		}
	}
	else {
		if ( channel < 13 || channel > 62 ) {
			freq = 0;
		}
		else {
			freq = 473 + (channel - 13) * 6;
		}
	}

	return freq;
}

static void usage(char *cmdname)
{
	fprintf(stderr,
		"ISDB-T recorder using DVB-T driver\n"
		"Usage: %s [options] [channel_ID] [duration] [output_file]\n"
		"  options:\n"
		"    -h, --help              show this message\n"
		"    -a, --adapter=NUM       adapter number (default: 0)\n"
		"    -c, --channel=ID        channel_ID\n"
		"    -f, --frequency=NUM     frequency for tuner in MHz\n"
		"    -d, --duration=NUM      recording time in seconds\n"
		"        --time=NUM          same as --duration\n"
		"    -l, --length=NUM        file size to record in MiB\n"
		"    -o, --output=FILE       file name to output\n"
		"        --tune-only         tune without output to file\n"
		"        --dvr-buffer=NUM    driver buffer size in MiB\n"
		"    -b, --b25               decrypt with ARIB B25\n"
		"    -v, --verbose           verbose message to <stderr>\n"
		"  channel_ID:\n"
		"    specify physical channel number\n"
		"    13 - 62 : UHF,  C13 - C63 : CATV\n"
		"  duration:\n"
		"    recording time in seconds, '-' for forever\n"
		"  output_file:\n"
		"    specify filename to output, '-' for <stdout>\n"
		, cmdname);
}

int main(int argc, char *argv[]) {
	int ret, c;
	int adapter = 0;
	int channel = 0;
	int catv = FALSE;
	int freq = 0;
	int duration = 0;
	int dmx_buffer_size = 0;
	int dvr_buffer_size = 0;
	int tune_only = FALSE;
	int verbose = FALSE;
	size_t length = 0;
	char *file = NULL;
	int long_index;
#ifdef B25
	int b25_flag = FALSE;
	ARIB_STD_B25 *b25;
	int b25_round = 4;
	int b25_strip = FALSE;
	int b25_emm = FALSE;
#endif
	dvb_control_t dvb;
	static struct option long_options[] = {
		{"adapter",    required_argument, NULL, 'a'},
		{"channel",    required_argument, NULL, 'c'},
		{"duration",   required_argument, NULL, 'd'},
		{"time",       required_argument, NULL, 'd'},
		{"frequency",  required_argument, NULL, 'f'},
		{"help",       no_argument,       NULL, 'h'},
		{"length",     required_argument, NULL, 'l'},
		{"output",     required_argument, NULL, 'o'},
		{"tune-only",  no_argument,       NULL, 't'},
		{"verbose",    no_argument,       NULL, 'v'},
		{"catv",       no_argument,       NULL, 'C'},
		{"CATV",       no_argument,       NULL, 'C'},
		{"uhf",        no_argument,       NULL, 'U'},
		{"UHF",        no_argument,       NULL, 'U'},
		{"dvr-buffer", required_argument, NULL, 'V'},
		{"dmx-buffer", required_argument, NULL, 'M'},
#ifdef B25
		{"b25",        no_argument,       NULL, 'b'},
		{"B25",        no_argument,       NULL, 'b'},
		{"b25-round",  required_argument, NULL, 'r'},
		{"b25-strip",  no_argument,       NULL, 's'},
		{"b25-emm",    no_argument,       NULL, 'm'},
#endif
		{NULL, 0, NULL, 0}
	};

	memset(&dvb, 0, sizeof(dvb));

	/* Parse options */
	while ( (c = getopt_long(argc, argv,
			"a:c:d:f:hl:o:v"
#ifdef B25
			"br:sm"
#endif
			, long_options, &long_index)) != -1 ) {
		switch ( c ) {
		  case 'a':
			adapter = strtol(optarg, NULL, 0);
			break;

		  case 'c':
			if ( toupper(*optarg) == 'C' ) {
				catv = TRUE;
				channel = strtol(optarg + 1, NULL, 0);
			}
			else {
				channel = strtol(optarg, NULL, 0);
			}
			break;

		  case 'd':
			if ( strcmp(optarg, "-") == 0 ) {
				/* forever */
				duration = -1;
			}
			else {
				duration = strtol(optarg, NULL, 0);
			}
			break;

		  case 'f':
			freq = strtol(optarg, NULL, 0);
			break;

		  case 'h':
			usage(argv[0]);
			exit(EXIT_SUCCESS);
			break;

		  case 'l':
			length = strtol(optarg, NULL, 0);
			break;

		  case 'o':
			file = optarg;
			break;

		  case 't':
			tune_only = TRUE;
			break;

		  case 'v':
			verbose = TRUE;
			break;

		  case 'C':
			catv = TRUE;
			break;

		  case 'U':
			catv = FALSE;
			break;

		  case 'V':
			dvr_buffer_size = strtol(optarg, NULL, 0);
			break;

		  case 'M':
			dmx_buffer_size = strtol(optarg, NULL, 0);
			break;

#ifdef B25
		  case 'b':
			b25_flag = TRUE;
			break;

		  case 'r':
			b25_round = strtol(optarg, NULL, 0);
			break;

		  case 's':
			b25_strip = TRUE;
			break;

		  case 'm':
			b25_emm = TRUE;
			break;
#endif

		  default:	/* ? */
			usage(argv[0]);
			exit(EXIT_FAILURE);
		} /* switch (c) */
	} /* while (getopt_long()) */

	/* Parse arguments */
	if ( channel == 0 && freq == 0 ) {
		if ( argc <= optind ) {
			usage(argv[0]);
			fprintf(stderr, "\n<channel_ID> is required.\n");
			exit(EXIT_FAILURE);
		}
		if ( toupper(*argv[optind]) == 'C' ) {
			catv = TRUE;
			channel = strtol(argv[optind] + 1, NULL, 0);
		}
		else {
			channel = strtol(argv[optind], NULL, 0);
		}
		optind++;
	}
	if ( channel == 0 && freq == 0 ) {
		usage(argv[0]);
		fprintf(stderr, "\n<channel_ID> is required.\n");
		exit(EXIT_FAILURE);
	}
	if ( channel != 0 && freq != 0 ) {
		usage(argv[0]);
		fprintf(stderr, "\n<channel> and <frequency> must be exclusive.\n");
		exit(EXIT_FAILURE);
	}
	if ( freq == 0 ) {
		freq = channel2freq(channel, catv);
		if ( freq <= 0 ) {
			fprintf(stderr, "Invalid channel ID: %s%d\n",
				(catv) ? "C" : "", channel);
			exit(EXIT_FAILURE);
		}
	}

	if ( duration == 0 && length == 0 ) {
		if ( argc <= optind ) {
			usage(argv[0]);
			fprintf(stderr, "\n<duration> is required.\n");
			exit(EXIT_FAILURE);
		}
		if ( strcmp(argv[optind], "-") != 0 ) {
			duration = strtol(argv[optind], NULL, 0);
		}
		optind++;
	}
	if ( duration < 0 ) {
		/* forever */
		duration = 0;
	}

	if ( !tune_only && file == NULL ) {
		if ( argc <= optind ) {
			usage(argv[0]);
			fprintf(stderr, "\n<output_file> is required.\n");
			exit(EXIT_FAILURE);
		}
		file = argv[optind];
	}

	/* Initialize DVB parameters */
	sprintf(dvb.fe_dev, "/dev/dvb/adapter%d/frontend0", adapter);
	sprintf(dvb.dmx_dev, "/dev/dvb/adapter%d/demux0", adapter);
	sprintf(dvb.dvr_dev, "/dev/dvb/adapter%d/dvr0", adapter);
	if ( verbose ) {
		dvb.flags |= FLAG_VERBOSE;
	}
	dvb.dvr_buffer_size = DVR_BUFFER_SIZE;
	if ( dvr_buffer_size > 0 ) {
		int r;
		dvb.dvr_buffer_size = dvr_buffer_size * 1024 * 1024;
		/* align to packet size */
		r = dvb.dvr_buffer_size % 188;
		if ( r != 0 ) {
			dvb.dvr_buffer_size += 188 - r;
		}
	}
	dvb.dmx_buffer_size = DMX_BUFFER_SIZE;
	if ( dmx_buffer_size > 0 ) {
		dvb.dmx_buffer_size = dmx_buffer_size * 1024 * 1024;
	}

#ifdef B25
	/* Initialize ARIB B25 library */
	if ( b25_flag ) {
		b25 = b25_init(b25_round, b25_strip, b25_emm);
		if ( b25 == NULL ) {
			fprintf(stderr, "Continue without ARIB B25.\n");
		}
		dvb.b25 = b25;
	}
#endif

	/* Initialize DVB device */
	ret = tune(&dvb, freq);
	if ( ret != 0 ) {
		goto done;
	}
	ret = dmx_route_dvr(&dvb);
	if ( ret != 0 ) {
		goto done;
	}

	/* main */
	if ( tune_only ) {
		struct sigaction sa;

		/* Set signal hander */
		memset(&sa, 0, sizeof(struct sigaction));
		sa.sa_handler = signal_handler;
		sa.sa_flags = SA_RESTART;
		sigaction(SIGINT,  &sa, NULL);
		sigaction(SIGTERM, &sa, NULL);
		sigaction(SIGALRM, &sa, NULL);

		/* wait for a signal */
		if ( duration > 0 ) {
			alarm(duration);
		}
		pause();

		/* Reset signal hander */
		memset(&sa, 0, sizeof(struct sigaction));
		sa.sa_handler = SIG_DFL;
		sigaction(SIGALRM, &sa, NULL);
		sigaction(SIGTERM, &sa, NULL);
		sigaction(SIGINT,  &sa, NULL);
	} /* tune only */
	else {
		if ( length > 0 ) {
			length *= 1024 * 1024;
		}
		ret = record_dvr(&dvb, file, duration, length);
	}

done:
	/* Close all the handlers and release ARIB B25 resources */
	dvb_shutdown(&dvb);

	return ret;
}
